"""An API client for the Open Trivia DB."""

##############################################################################
# Local imports.
from .category import Category
from .client import OpenTriviaDB
from .counts import Counts
from .question import Difficulty, Question, Type
from .response import (
    Code,
    InvalidParameter,
    NoResults,
    RateLimit,
    TokenEmpty,
    TokenNotFound,
)

##############################################################################
# Exports.
__all__ = [
    "Category",
    "Code",
    "Counts",
    "Difficulty",
    "InvalidParameter",
    "NoResults",
    "OpenTriviaDB",
    "Question",
    "RateLimit",
    "TokenEmpty",
    "TokenNotFound",
    "Type",
]

### __init__.py ends here



"""Provides a widget for displaying an answer."""

##############################################################################
# Textual imports.
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Digits, Label


##############################################################################
class Answer(Horizontal):
    """Display an answer to a question."""

    DEFAULT_CSS = """
    Answer {
        width: 1fr;
        height: auto;
        padding: 0 1 0 1;

        Digits {
            width: auto;
            height: auto;
            padding-right: 1;
            color: $success;
        }

        Label {
            width: 1fr;
            height: auto;
            min-height: 3;
            content-align: left middle;
        }
    }
    """

    def __init__(self, answer: str, number: int, classes: str | None = None) -> None:
        """Initialise the answer.

        Args:
            answer: The answer text.
            number: The number the user will use to select the answer.
            classes: The classes to apply to the widget.
        """
        super().__init__(id=f"answer-{number}", classes=classes)
        self._answer = answer
        """The answer to show."""
        self._number = number
        """The number of the answer."""

    def compose(self) -> ComposeResult:
        """Compose the answer widget's content."""
        yield Digits(f"{self._number}:")
        yield Label(self._answer)


### answer.py ends here



"""Provides the screen used to take a quiz."""

##############################################################################
# Rich imports.
from rich.console import Group
from rich.table import Table

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Center, Grid, Vertical
from textual.css.query import NoMatches
from textual.reactive import var
from textual.screen import ModalScreen
from textual.widgets import Button, Digits, Label, LoadingIndicator, OptionList

##############################################################################
# Textual countdown imports.
from textual_countdown import Countdown

##############################################################################
# Local imports.
from ...opentdb import OpenTriviaDB, Question
from ..data.quiz_parameters import QuizParameters, QuizTimer
from ..widgets import Answers


##############################################################################
class QuizTaker(ModalScreen[None]):
    """Screen for taking a quiz."""

    DEFAULT_CSS = """
    QuizTaker {
        align: center middle;

        .hidden {
            display: none;
        }

        &> Vertical {
            background: $surface;
            border: panel $primary;
            border-title-color: $accent;
        }

        #loader {
            width: 30%;
            min-width: 50;
            height: auto;
            LoadingIndicator {
               height: 3;
            }
        }

        #confirmer {
            padding: 1 1 0 1;
            width: auto;
            height: auto;
            Label {
                width: auto;
                margin-bottom: 1;
                text-align: center;
            }
            Center {
                width: 100%;
                layout: horizontal;
            }
            Button {
                margin-right: 1;
            }
        }

        #taker {
            width: auto;
            max-width: 60;
            height: auto;

            Grid {
                grid-size: 4 2;
                height: auto;
                grid-rows: auto;
                margin-top: 1;
                width: 100%;

                Label, Digits {
                    width: 1fr;
                    text-align: center;
                }
            }

            Countdown {
                width: 100%;
            }

            #question-text {
                width: 1fr;
                height: auto;
                padding: 1 2 1 2;
                text-style: bold;
            }

            #answers {
                width: 1fr;
                border: none;
                margin: 0 1 1 1;
                padding: 1;
                &:focus {
                    border: none;
                }
            }
        }

        #results {
            width: auto;
            max-width: 60%;
            height: auto;
            max-height: 80%;
            padding: 1 2 0 2;
            Label {
                width: auto;
            }
            #answers {
                height: 1fr;
                margin: 1 0 1 0;
                &> .option-list--option {
                    padding: 0 5 1 0;
                }
            }
            Center {
                width: 100%;
                layout: horizontal;
            }
        }
    }
    """

    _question: var[int] = var(-1, init=False)
    """The number of the question being asked."""

    _answers: var[list[str | None]] = var(list, init=False)
    """Holds the answers given."""

    _correct: var[int] = var(0, init=False)
    """The count of correct answers."""

    _wrong: var[int] = var(0, init=False)
    """The count of wrong answers."""

    def __init__(self, client: OpenTriviaDB, quiz: QuizParameters) -> None:
        """Initialise the quiz taking screen.

        Args:
            client: The API client.
            quiz: The parameters for the quiz to take.
        """
        super().__init__()
        self._client = client
        """The trivia database client object."""
        self._quiz_parameters = quiz
        """The parameters of the quiz we're going to take."""
        self._quiz: list[Question] = []
        """The questions being asked as part of the quiz."""

    def compose(self) -> ComposeResult:
        """Compose the quiz taking screen."""

        with Vertical(id="loader") as quiz_loader:
            quiz_loader.border_title = f"Loading: {self._quiz_parameters.title}"
            yield LoadingIndicator()
            with Center():
                yield Button("Cancel", id="cancel")

        with Vertical(id="confirmer", classes="hidden") as quiz_confirmer:
            quiz_confirmer.border_title = (
                f"Ready to take '{self._quiz_parameters.title}'?"
            )
            yield Label("The quiz is now loaded and ready; are you all set to take it?")
            with Center():
                yield Label(id="details")
            with Center():
                yield Button("Let's go!", id="take", variant="success")
                yield Button("Cancel", id="cancel", variant="error")

        with Vertical(id="taker", classes="hidden") as quiz_taker:
            quiz_taker.border_title = self._quiz_parameters.title
            with Grid():
                yield Label("Question")
                yield Label("Out Of")
                yield Label("Correct")
                yield Label("Wrong")
                yield Digits("0", id="question")
                yield Digits("0", id="out-of")
                yield Digits("0", id="correct")
                yield Digits("0", id="wrong")
            if self._quiz_parameters.timer_type != QuizTimer.NONE:
                yield Countdown()
            yield Label(id="question-text")
            yield Answers()
            with Center():
                yield Button("Retire", id="cancel")

        with Vertical(id="results", classes="hidden") as quiz_results:
            quiz_results.border_title = f"Results of '{self._quiz_parameters.title}'"
            yield Label(id="final-score")
            yield OptionList(id="answers")
            with Center():
                yield Button("Close", id="cancel")

    def on_mount(self) -> None:
        """Start the process of loading up the quiz once the DOM is ready."""
        self._load_quiz()

    async def _quiz_summary(self) -> str:
        """A descriptive summary of the quiz."""
        summary = f"{len(self._quiz)} question{'' if len(self._quiz) == 1 else 's'}\n"
        if self._quiz_parameters.category is None:
            summary += "Any category\n"
        else:
            summary += f"{(await self._client.category(self._quiz_parameters.category)).name}\n"
        if self._quiz_parameters.difficulty is None:
            summary += "Any difficulty level\n"
        else:
            summary += (
                f"{self._quiz_parameters.difficulty.capitalize()} difficulty level\n"
            )
        if self._quiz_parameters.question_type is None:
            summary += "Multiple choice and true/false questions\n"
        else:
            summary += (
                f"{'True or false' if self._quiz_parameters.question_type == 'boolean' else 'Multiple choice'}"
                " questions\n"
            )
        if self._quiz_parameters.timer_type == QuizTimer.PER_QUESTION:
            summary += f"{self._quiz_parameters.timer_value} seconds per question\n"
        if self._quiz_parameters.timer_type == QuizTimer.WHOLE_QUIZ:
            summary += (
                f"{self._quiz_parameters.timer_value} seconds to complete the quiz\n"
            )
        return summary

    @work
    async def _load_quiz(self) -> None:
        """Load the quiz from the backend."""
        self._quiz = await self._client.questions(
            amount=self._quiz_parameters.number_of_questions,
            category=self._quiz_parameters.category,
            difficulty=self._quiz_parameters.difficulty,
            of_type=self._quiz_parameters.question_type,
        )
        with self.app.batch_update():
            self.query_one("#loader").set_class(True, "hidden")
            self.query_one("#confirmer").set_class(False, "hidden")
            self.query_one("#confirmer #details", Label).update(
                await self._quiz_summary()
            )
            self.query_one("#take").focus()

    async def _watch__question(self) -> None:
        """React to the question number being bumped."""
        self.query_one("#question", Digits).update(str(self._question + 1))
        self.query_one("#question-text", Label).update(
            self._quiz[self._question].question
        )
        self.query_one(Answers).question = self._quiz[self._question]
        if self._quiz_parameters.timer_type == QuizTimer.PER_QUESTION:
            self.query_one(Countdown).start(self._quiz_parameters.timer_value)

    def _watch__correct(self) -> None:
        """React to the right count being changed."""
        self.query_one("#correct", Digits).update(str(self._correct))

    def _watch__wrong(self) -> None:
        """React to the wrong count being changed."""
        self.query_one("#wrong", Digits).update(str(self._wrong))

    @on(Button.Pressed, "#cancel")
    def cancel_quiz(self) -> None:
        """Cancel the current quiz."""
        self.dismiss()

    @on(Button.Pressed, "#take")
    def take_quiz(self) -> None:
        """Start the process of taking the quiz."""
        with self.app.batch_update():
            self.query_one("#confirmer").set_class(True, "hidden")
            self.query_one("#taker").set_class(False, "hidden")
            self.query_one("#out-of", Digits).update(str(len(self._quiz)))
            self._question = 0
            self.query_one(Answers).focus()
            if self._quiz_parameters.timer_type == QuizTimer.WHOLE_QUIZ:
                self.query_one(Countdown).start(self._quiz_parameters.timer_value)

    @on(Answers.Correct)
    @on(Answers.Incorrect)
    def process_answer(self, event: Answers.Correct | Answers.Incorrect) -> None:
        """Process an answer given to the Answers widget.

        Args:
            event: The event to process.
        """
        if isinstance(event, Answers.Correct):
            self._correct += 1
        else:
            self._wrong += 1
        self._answers.append(event.answer)
        if self._question < (len(self._quiz) - 1):
            self._question += 1
        else:
            self.call_next(self._show_result)

    @on(Countdown.Finished)
    def process_timeout(self) -> None:
        """Process a timeout while answering a question."""
        if self._quiz_parameters.timer_type == QuizTimer.PER_QUESTION:
            self.query_one(Answers).skip_question()
        else:
            questions_left = len(self._quiz) - len(self._answers)
            self._answers.extend([None] * questions_left)
            self._wrong += questions_left
            self.call_next(self._show_result)

    def _question_result(self, question: int) -> Group:
        """Get the result display for a question.

        Args:
            question: The question to get the display for.

        Returns:
            The display.
        """
        question_row = Table.grid()
        question_row.add_column(width=5, justify="right")
        question_row.add_column(width=1)
        question_row.add_column(ratio=1)
        question_row.add_row(f"{question + 1}. ", " ", self._quiz[question].question)
        answer_row = Table.grid()
        answer_row.add_column(width=6)
        answer_row.add_column(ratio=1)
        colour = (
            "green"
            if self._answers[question] == self._quiz[question].correct_answer
            else "red"
        )
        answer_row.add_row(
            "", f"[{colour}]{self._answers[question] or '(No answer given)'}[/]"
        )
        return Group(question_row, answer_row)

    async def _show_result(self) -> None:
        """Show the result of the quiz."""
        # Ensure that any running countdown is cancelled; we're all done
        # here.
        try:
            self.query_one(Countdown).cancel()
        except NoMatches:
            # Failure to find a countdown is fine; it just means there's no
            # countdown to begin with.
            pass
        # Having done that, show the results of the quiz.
        with self.app.batch_update():
            self.query_one("#taker").set_class(True, "hidden")
            self.query_one("#results").set_class(False, "hidden")
            self.query_one("#final-score", Label).update(
                f"Final score is {self._correct} out of {len(self._quiz)}."
            )
            self.query_one("#answers", OptionList).add_options(
                self._question_result(number) for number in range(len(self._answers))
            )
            self.query_one("#results #cancel").focus()


### quiz_taker.py ends here



"""Modal dialog for getting the parameters to make a quiz."""

##############################################################################
# Python imports.
from typing import cast, get_args

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.validation import Integer, Length
from textual.widgets import Button, Input, Label, Select

##############################################################################
# Local imports.
from ...opentdb import Category, Difficulty, Type
from ..data import QuizParameters, QuizTimer


##############################################################################
class QuizMaker(ModalScreen[QuizParameters | None]):
    """Modal screen for prompting the user for quiz parameters."""

    DEFAULT_CSS = """
    QuizMaker {
        align: center middle;

        &> Vertical {
            width: 70%;
            height: auto;
            background: $surface;
            border: panel $primary;
            border-title-color: $accent;

            Label {
                margin: 1 0 0 1;
                &:disabled {
                    color: $text-muted;
                }
            }

            &> Horizontal {
                margin-top: 1;
                height: auto;
                align: right middle;
                Button {
                    margin-left: 1;
                }
            }
        }
    }
    """

    BINDINGS = [("escape", "cancel"), ("f2", "okay")]

    def __init__(
        self, categories: list[Category], quiz: QuizParameters | None = None
    ) -> None:
        """Initialise the quiz maker.

        Args:
            categories: The list of categories that the user can select from.
            quiz: An optional existing quiz to populate the dialog with.
        """
        super().__init__()
        self._categories = categories
        """The known categories for the questions."""
        self._quiz = quiz
        """Existing quiz parameters to populate the list with."""

    def compose(self) -> ComposeResult:
        """Compose the content of the dialog."""
        with Vertical() as dialog:
            dialog.border_title = "Quiz Options"
            yield Label("Title:")
            yield Input(
                placeholder="The title for the quiz",
                validators=Length(minimum=1),
                id="title",
            )
            yield Label("Number of questions:")
            yield Input(
                "10",
                placeholder="Number of questions to ask",
                type="integer",
                validators=Integer(minimum=1),
                id="number",
            )
            yield Label("Category:")
            yield Select[int](
                ((category.name, category.id) for category in self._categories),
                prompt="Any Category",
                id="category",
            )
            yield Label("Difficulty:")
            yield Select[str](
                (
                    (difficulty.capitalize(), difficulty)
                    for difficulty in get_args(Difficulty)
                ),
                prompt="Any Difficulty",
                id="difficulty",
            )
            yield Label("Question Type:")
            yield Select[str](
                (
                    (
                        "Multiple Choice"
                        if question_type == "multiple"
                        else "True or False",
                        question_type,
                    )
                    for question_type in get_args(Type)
                ),
                prompt="Any Type",
                id="type",
            )
            yield Label("Timer Type:")
            yield Select[QuizTimer](
                ((timer.description, timer) for timer in list(QuizTimer)),
                value=QuizTimer.NONE,
                allow_blank=False,
                id="timer-type",
            )
            yield Label("Timer Seconds:", classes="timer", disabled=True)
            yield Input(
                "30",
                placeholder="Number of seconds for the timer",
                type="integer",
                validators=Integer(minimum=1),
                id="timer-seconds",
                classes="timer",
                disabled=True,
            )
            with Horizontal():
                yield Button("Okay [dim]\\[F2][/]", id="okay")
                yield Button("Cancel [dim]\\[Esc][/]", id="cancel")

    def on_mount(self) -> None:
        """Configure the dialog once the DOM is ready."""
        if self._quiz is not None:
            self.query_one("#title", Input).value = self._quiz.title
            self.query_one("#number", Input).value = str(self._quiz.number_of_questions)
            self.query_one("#category", Select).value = (
                self._quiz.category if self._quiz.category else Select.BLANK
            )
            self.query_one("#difficulty", Select).value = (
                self._quiz.difficulty or Select.BLANK
            )
            self.query_one("#type", Select).value = (
                self._quiz.question_type or Select.BLANK
            )
            self.query_one("#timer-type", Select).value = self._quiz.timer_type
            self.query_one("#timer-seconds", Input).value = str(self._quiz.timer_value)
            self._update_timer_fields()

    @on(Select.Changed, "#timer-type")
    def _update_timer_fields(self) -> None:
        """Refresh the state of the timer input fields."""
        timer_type = self.query_one("#timer-type", Select).value
        for widget in self.query(".timer").results():
            widget.disabled = timer_type == QuizTimer.NONE
            if isinstance(widget, Input):
                if widget.disabled:
                    widget.validators = []
                else:
                    widget.validators = [Integer(minimum=1)]
                widget.validate(widget.value)

    @on(Button.Pressed, "#okay")
    def action_okay(self) -> None:
        """React to the user confirming their choices."""
        okay = True
        if not self.query_one("#title", Input).is_valid:
            self.notify("Please enter a title", title="Missing Title", severity="error")
            okay = False
        if not self.query_one("#number", Input).is_valid:
            self.notify(
                "Please enter a valid number", title="Invalid Number", severity="error"
            )
            okay = False
        timer_seconds = self.query_one("#timer-seconds", Input)
        if not timer_seconds.disabled:
            if not timer_seconds.is_valid:
                self.notify(
                    "Please enter a valid number of seconds for the timer",
                    title="Invalid Timer",
                    severity="error",
                )
                okay = False
        if okay:
            self.dismiss(
                QuizParameters(
                    title=self.query_one("#title", Input).value,
                    number_of_questions=int(self.query_one("#number", Input).value),
                    category=category
                    if isinstance(
                        category := self.query_one("#category", Select).value, int
                    )
                    else None,
                    difficulty=(
                        cast(Difficulty, difficulty)
                        if isinstance(
                            difficulty := self.query_one("#difficulty", Select).value,
                            str,
                        )
                        else None
                    ),
                    question_type=(
                        cast(Type, question_type)
                        if isinstance(
                            question_type := self.query_one("#type", Select).value, str
                        )
                        else None
                    ),
                    timer_type=(
                        timer_type
                        if isinstance(
                            timer_type := self.query_one("#timer-type", Select).value,
                            QuizTimer,
                        )
                        else QuizTimer.NONE
                    ),
                    timer_value=int(
                        self.query_one("#timer-seconds", Input).value or "0"
                    ),
                )
            )

    @on(Button.Pressed, "#cancel")
    def action_cancel(self) -> None:
        """React to the user cancelling their choices."""
        self.dismiss(None)


### quiz_maker.py ends here



"""Provides a class that holds questions pulled from the API."""

##############################################################################
# Python imports.
from dataclasses import dataclass, field
from html import unescape
from typing_extensions import Literal

##############################################################################
Type = Literal["multiple", "boolean"]
"""The valid types of question."""

##############################################################################
Difficulty = Literal["easy", "medium", "hard"]
"""The valid question difficulty levels."""


##############################################################################
@dataclass
class Question:
    """Holds the details of a question."""

    type: Type
    """The type of the question."""

    difficulty: Difficulty
    """The difficulty level of the question."""

    category: str
    """The name of the category that the question is within."""

    question: str
    """The question itself."""

    correct_answer: str
    """The correct answer to the question."""

    incorrect_answers: list[str] = field(default_factory=list)
    """The incorrect answers for the question."""

    def __post_init__(self) -> None:
        """Tidy up the question data once it's been loaded up."""
        self.category = unescape(self.category)
        self.question = unescape(self.question)
        self.correct_answer = unescape(self.correct_answer)
        self.incorrect_answers = [unescape(answer) for answer in self.incorrect_answers]

    @property
    def answers(self) -> list[str]:
        """All of the answers to the question, sorted in alphabetical order."""
        return sorted([self.correct_answer, *self.incorrect_answers], key=str.casefold)


### question.py ends here



"""Provides a widget for displaying a set of answers."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from functools import partial

##############################################################################
# Textual imports.
from textual.containers import Vertical
from textual.message import Message
from textual.reactive import var

##############################################################################
# Local imports.
from ...opentdb import Question
from .answer import Answer


##############################################################################
class Answers(Vertical, can_focus=True):
    """Displays and a collection of answers and prompts for one."""

    DEFAULT_CSS = """
    Answers {
        height: auto;
        opacity: 0.5;
        border-top: dashed $primary;
        border-bottom: dashed $primary;
        margin: 0 2 1 2;

        &:focus {
            opacity: 1;
            background: $boost;
        }

        Answer.sep {
            border-top: dashed $primary;
        }

        Answer.correct * {
            background: $success 50%;
        }

        Answer.incorrect * {
            background: $error 50%;
        }
    }
    """

    BINDINGS = [((str(n), f"answer({n})")) for n in range(1, 5)]

    question: var[Question | None] = var(None, init=False)
    """The question whose answers are to be shown."""

    @dataclass
    class Given(Message):
        """Base class for the given answer."""

        answer: str | None = None
        """The answer that was given."""

    class Correct(Given):
        """Message sent when the player gives a correct answer."""

    class Incorrect(Given):
        """message sent when the player gives an incorrect answer."""

    answer: var[Given | None] = var(None)
    """The answer being processed."""

    async def _watch_question(self) -> None:
        """React to a new question being set."""
        await self.query(Answer).remove()
        if self.question is not None:
            await self.mount_all(
                Answer(question, number + 1, classes=("sep" if number else ""))
                for number, question in enumerate(self.question.answers)
            )
            self.answer = None

    def action_answer(self, answer: int) -> None:
        """Process a player's answer.

        Args:
            answer: The answer given by the player.
        """
        if self.question is None or self.answer is not None:
            return
        self.answer = (
            self.Correct
            if self.question.correct_answer == self.question.answers[answer - 1]
            else self.Incorrect
        )(self.question.answers[answer - 1])
        self.query_one(f"#answer-{answer}").set_class(
            True, "correct" if isinstance(self.answer, self.Correct) else "incorrect"
        )
        self.set_timer(0.6, partial(self.post_message, self.answer))

    def skip_question(self) -> None:
        """Skip the current question."""
        self.query(Answer).set_class(True, "incorrect")
        self.set_timer(0.6, partial(self.post_message, self.Incorrect()))


### answers.py ends here



"""Provides the application's custom widgets."""

##############################################################################
# Local imports.
from .answers import Answers
from .logo import Logo
from .question_counts import QuestionCounts
from .quiz_list import QuizList

##############################################################################
# Exports.
__all__ = ["Answers", "Logo", "QuestionCounts", "QuizList"]

### __init__.py ends here



"""Provides a widget that holds the list of saved quiz parameters."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from dataclasses import dataclass
from json import dumps, loads

##############################################################################
# Textual imports.
from textual import on
from textual.events import Mount
from textual.message import Message
from textual.reactive import var
from textual.widgets import OptionList

##############################################################################
# Backward-compatible typing.
from typing_extensions import Self

##############################################################################
# Local imports.
from ..data import QuizParameters, quizzes_file


##############################################################################
class QuizList(OptionList):
    """Holds the list of available quizzes."""

    DEFAULT_CSS = """
    QuizList {
        height: 1fr;
        border: none;

        &:focus {
            border: none;
            background: $boost;
        }
    }
    """

    BINDINGS = [
        ("s, j, right", "cursor_down"),
        ("w, k, left", "cursor_up"),
    ]

    quizzes: var[list[QuizParameters]] = var(list, always_update=True, init=False)
    """The list of all the quiz parameters."""

    def clear_options(self) -> Self:
        """Workaround for https://github.com/Textualize/textual/issues/3714"""
        super().clear_options()
        self._clear_content_tracking()
        return self

    @dataclass
    class Changed(Message):
        """Message sent when the quiz list has been changed."""

        quiz_list: QuizList
        """The quiz list that was changed."""

    def _changed(self) -> None:
        """Handle the list being changed."""
        quizzes_file().write_text(
            dumps(
                [quiz.as_json for quiz in self.quizzes],
                indent=4,
                default=lambda x: x.value,
            ),
            encoding="utf-8",
        )
        self.post_message(self.Changed(self))

    @on(Mount)
    def _load(self) -> None:
        """Load the current quiz list."""
        try:
            self.quizzes = [
                QuizParameters(**quiz)
                for quiz in loads(quizzes_file().read_text(encoding="utf-8"))
            ]
        except FileNotFoundError:
            pass
        else:
            self.post_message(self.Changed(self))

    def _watch_quizzes(self) -> None:
        """Refresh the prompt list when the quiz list is reassigned."""
        self.clear_options().add_options([quiz.title for quiz in self.quizzes])
        if self.quizzes:
            self.highlighted = 0

    def add_quiz(self, quiz: QuizParameters) -> None:
        """Add a new quiz to the list.

        Args:
            quiz: The quiz parameters to add.
        """
        self.add_option(quiz.title)
        self.quizzes.append(quiz)
        if self.highlighted is None:
            self.highlighted = 0
        self._changed()

    def modify_quiz(self, index: int, quiz: QuizParameters) -> None:
        """Modify the given quiz with the new parameters.

        Args:
            index: The index of the quiz to update.
            quiz: The quiz parameters to replace the quiz with.
        """
        self.quizzes[index] = quiz
        self.replace_option_prompt_at_index(index, quiz.title)
        self._changed()

    def remove_quiz(self, index: int) -> None:
        """Remove the quiz at the given index.

        Args:
            index: The index of the quiz to remove.
        """
        del self.quizzes[index]
        self.remove_option_at_index(index)
        self._changed()


### quiz_list.py ends here



"""The main screen of the application."""

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.screen import Screen
from textual.widgets import Button, Label

##############################################################################
# Local imports.
from ... import __version__
from ...opentdb import OpenTriviaDB
from ..widgets import Logo, QuestionCounts, QuizList
from .confirm import Confirm
from .quiz_maker import QuizMaker
from .quiz_taker import QuizTaker


##############################################################################
class Main(Screen[None]):
    """The main screen of the application."""

    CSS = """
    #version {
        width: 1fr;
        content-align: center middle;
        background: $primary;
        padding-bottom: 1;
    }

    QuizList {
        border-top: solid $boost;
        background: $primary;
    }

    #buttons {
        height: auto;
        background: $primary;
        border-top: solid $boost;
        padding-bottom: 1;
        Button {
            width: 1fr;
            min-width: 0;
            margin: 0 1 0 1;
        }
    }
    """

    BINDINGS = [
        ("r", "run"),
        ("n", "new"),
        ("e", "edit"),
        ("d", "delete"),
        ("q", "quit"),
    ]

    def __init__(self) -> None:
        super().__init__()
        self._trivia = OpenTriviaDB()
        """The Open Trivia DB client."""

    def compose(self) -> ComposeResult:
        """Compose the main screen."""
        yield Logo()
        yield Label(f"v{__version__}", id="version")
        yield QuestionCounts("All Questions")
        yield QuizList()
        with Horizontal(id="buttons"):
            yield Button("Run [dim]\\[r][/]", id="run", disabled=True)
            yield Button("New [dim]\\[n][/]", id="new")
            yield Button("Edit [dim]\\[e][/]", id="edit", disabled=True)
            yield Button("Delete [dim]\\[d][/]", id="delete", disabled=True)
            yield Button("Quit [dim]\\[q][/]", id="quit")

    def on_mount(self) -> None:
        """Load up the data for the main display."""
        self._load_counts()

    def _connect_error(self) -> None:
        """Let the user know there was an error when connecting to the backend."""
        self.notify(
            "Unable to connect to https://opentdb.com/ at the moment!",
            severity="error",
            timeout=8,
        )
        self.query_one("#buttons").disabled = True

    @work
    async def _load_counts(self) -> None:
        """Load up the question counts."""
        try:
            self.query_one(QuestionCounts).counts = await self._trivia.overall_counts()
        except self._trivia.RequestError:
            self.query_one(QuestionCounts).counts = QuestionCounts.Unavailable()
            self._connect_error()

    @on(QuizList.Changed)
    def _update_buttons(self, event: QuizList.Changed) -> None:
        """Update the state of the buttons.

        Args:
            event: The event to handle.
        """
        self.query_one("#run").disabled = event.quiz_list.highlighted is None
        self.query_one("#edit").disabled = event.quiz_list.highlighted is None
        self.query_one("#delete").disabled = event.quiz_list.highlighted is None

    @on(Button.Pressed, "#run")
    def action_run(self) -> None:
        """Run the currently-hilighted quiz."""
        quizzes = self.query_one(QuizList)
        if quizzes.highlighted is not None:
            self.app.push_screen(
                QuizTaker(self._trivia, quizzes.quizzes[quizzes.highlighted])
            )

    @on(Button.Pressed, "#new")
    @work
    async def action_new(self) -> None:
        """Create a new quiz."""
        try:
            if quiz := await self.app.push_screen_wait(
                QuizMaker(await self._trivia.categories())
            ):
                self.query_one(QuizList).add_quiz(quiz)
        except self._trivia.RequestError:
            self._connect_error()

    @on(Button.Pressed, "#edit")
    @work
    async def action_edit(self) -> None:
        """Edit the currently-highlighted quiz."""
        quizzes = self.query_one(QuizList)
        if (to_edit := quizzes.highlighted) is not None:
            try:
                if quiz := await self.app.push_screen_wait(
                    QuizMaker(await self._trivia.categories(), quizzes.quizzes[to_edit])
                ):
                    quizzes.modify_quiz(to_edit, quiz)
            except self._trivia.RequestError:
                self._connect_error()

    @on(Button.Pressed, "#delete")
    @work
    async def action_delete(self) -> None:
        """Delete the currently-highlighted quiz."""
        if (to_delete := self.query_one(QuizList).highlighted) is not None:
            if await self.app.push_screen_wait(
                Confirm("Delete Quiz", "Are you sure you want to delete that quiz?")
            ):
                self.query_one(QuizList).remove_quiz(to_delete)

    @on(Button.Pressed, "#quit")
    def action_quit(self) -> None:
        """Quit the application."""
        self.app.exit()


### main.py ends here



"""Code relating to the data stored by the application."""

##############################################################################
# Local imports.
from .locations import data_dir
from .quiz_parameters import QuizParameters, QuizTimer
from .quizzes_file import quizzes_file

##############################################################################
# Exports.
__all__ = ["data_dir", "QuizParameters", "quizzes_file", "QuizTimer"]

### __init__.py ends here



"""Provides code for working with quiz parameters."""

##############################################################################
# Python imports.
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, cast

##############################################################################
# Local imports.
from ...opentdb import Difficulty, Type


##############################################################################
class QuizTimer(Enum):
    """The types of timer a quiz can have."""

    NONE = 0
    """No timer."""

    PER_QUESTION = 1
    """Timer per question."""

    WHOLE_QUIZ = 2
    """Timer for whole quiz."""

    @property
    def description(self) -> str:
        """The description of the timer type."""
        return cast(
            dict[QuizTimer, str],
            {
                self.NONE: "No Timer",
                self.PER_QUESTION: "Seconds per question",
                self.WHOLE_QUIZ: "Seconds for the whole quiz",
            },
        )[self]


##############################################################################
@dataclass
class QuizParameters:
    """Holds the chosen parameters for making a quiz."""

    title: str
    """The title for the quiz."""

    number_of_questions: int = 10
    """The number of questions in the quiz."""

    category: int | None = None
    """The category for the quiz."""

    difficulty: Difficulty | None = None
    """The difficulty for the quiz."""

    question_type: Type | None = None
    """The type of question to ask."""

    timer_type: QuizTimer = QuizTimer.NONE
    """The type of timer to use for the quiz."""

    timer_value: int = 0
    """The number of seconds for the timer, if there is one."""

    def __post_init__(self) -> None:
        """Tidy up the data post-init."""
        # If we've been pulled in from a JSON source, it's likely that we
        # got the timer type back as an integer; to save a lot of cocking
        # about with decoders and stuff, let's just cast it here.
        self.timer_type = QuizTimer(self.timer_type)

    @property
    def as_json(self) -> dict[str, Any]:
        """The quiz parameters as a JSON-friendly dictionary."""
        return asdict(self)


### quiz_parameters.py ends here



"""Provides code for dealing with response values."""

##############################################################################
# Python imports.
from enum import Enum


##############################################################################
class NoResults(Exception):
    """Exception raised if there are no results."""


##############################################################################
class InvalidParameter(Exception):
    """Exception raised if there is an invalid parameter."""


##############################################################################
class TokenNotFound(Exception):
    """Exception raised if the passed token isn't found."""


##############################################################################
class TokenEmpty(Exception):
    """Exception raised if the given token has used all possible questions."""


##############################################################################
class RateLimit(Exception):
    """Exception raised if there have been too many requests."""


##############################################################################
class Code(Enum):
    """The response codes that the trivia API can send back."""

    UNKNOWN = -1
    """The response code is so far unknown."""

    SUCCESS = 0
    """Returned results successfully."""

    NO_RESULTS = 1
    """Could not return results.

    The API doesn't have enough questions for your query. (Ex. Asking
    for 50 Questions in a Category that only has 20.)
    """

    INVALID_PARAMETER = 2
    """Contains an invalid parameter.

    Arguments passed in aren't valid. (Ex. Amount = Five)
    """

    TOKEN_NOT_FOUND = 3
    """Session Token does not exist."""

    TOKEN_EMPTY = 4
    """Session Token has returned all possible questions for the specified query.

    Resetting the Token is necessary.
    """

    RATE_LIMIT = 5
    """Too many requests have occurred.

    Each IP can only access the API once every 5 seconds.
    """

    def maybe_raise(self) -> None:
        """Raise any appropriate exceptions based on the code."""
        if self is Code.SUCCESS:
            return
        try:
            raise {
                Code.NO_RESULTS: NoResults,
                Code.INVALID_PARAMETER: InvalidParameter,
                Code.TOKEN_NOT_FOUND: TokenNotFound,
                Code.TOKEN_EMPTY: TokenEmpty,
                Code.RATE_LIMIT: RateLimit,
            }[self]()
        except KeyError:
            raise ValueError(f"Not a valid response code: {self}") from None


### response.py ends here



"""Provides a class that holds details of a quiz category."""

##############################################################################
# Python imports.
from dataclasses import dataclass
from html import unescape


##############################################################################
@dataclass
class Category:
    """Class that holds the details of a quiz category."""

    id: int
    """The ID of the category"""

    name: str
    """The name of the category."""

    def __post_init__(self) -> None:
        """Tidy up the question data once it's been loaded up."""
        self.name = unescape(self.name)


### category.py ends here



"""Provides a widget that shows question count totals."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Humanize imports.
from humanize import intcomma

##############################################################################
# Rich imports.
from rich.align import Align
from rich.console import Group
from rich.table import Table

##############################################################################
# Textual imports.
from textual.app import RenderResult
from textual.reactive import reactive
from textual.widget import Widget

##############################################################################
# Local imports.
from ...opentdb import Counts


##############################################################################
class QuestionCounts(Widget):
    """Widget that shows question counts."""

    DEFAULT_CSS = """
    QuestionCounts {
        height: auto;
        width: 1fr;
        background: $primary;
        color: $text-muted;
    }
    """

    class Unavailable:
        """Class used to mark that counts are unavailable."""

    counts: reactive[Counts | None | Unavailable] = reactive(None)
    """The counts to show."""

    def __init__(  # pylint:disable=redefined-builtin
        self,
        title: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Initialise the question counts widget.

        Args:
            title: The title of the counts to show.
            name: The name of the question counts widget.
            id: The ID of the question counts widget in the DOM.
            classes: The CSS classes of the question counts widget.
            disabled: Whether the question counts widget is disabled or not.
        """
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self._title = title
        """The title of the counts."""

    def render(self) -> RenderResult:
        """Render the content of the widget."""
        display = Table(expand=True)
        for title in ("Total", "Pending", "Verified", "Rejected"):
            display.add_column(title, no_wrap=True, justify="right", ratio=1)
        if self.counts is None:
            display.add_row(*(["Loading..."] * 4))
        elif isinstance(self.counts, self.Unavailable):
            display.add_row(*(["[red]Unavailable[/]"] * 4))
        else:
            display.add_row(
                intcomma(self.counts.questions),
                intcomma(self.counts.pending),
                intcomma(self.counts.verified),
                intcomma(self.counts.rejected),
            )
        return Group(Align(self._title, align="center"), display)


### question_counts.py ends here



"""Provides a class that holds counts."""

##############################################################################
# Python imports.
from dataclasses import dataclass


##############################################################################
@dataclass
class Counts:
    """Class that holds the counts of questions."""

    questions: int
    """The total number of questions."""

    pending: int
    """The number of pending questions."""

    verified: int
    """The number of verified questions."""

    rejected: int
    """The number of rejected questions."""


### counts.py ends here



"""The Open Trivia DB client."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from json import loads
from typing import Final

##############################################################################
# HTTPX imports.
from httpx import AsyncClient, HTTPStatusError, RequestError

##############################################################################
# Local imports.
from .category import Category
from .counts import Counts
from .question import Difficulty, Question, Type
from .response import Code


##############################################################################
class OpenTriviaDB:
    """Client for the Open Trivia DB API."""

    AGENT: Final[str] = "Quizzical (https://github.com/davep/quizzical)"
    """The agent string to use when talking to the API."""

    _BASE: Final[str] = "https://opentdb.com/"
    """The base of the URL for the API."""

    class RequestError(Exception):
        """Exception raised if there was a problem making an API request."""

    def __init__(self) -> None:
        """Initialise the API client object."""
        self._client_: AsyncClient | None = None
        """The HTTPX client."""
        self._categories: list[Category] = []
        """The list of categories."""
        self._overall_counts: Counts | None = None
        """The overall counts of questions in the backend."""
        self._category_counts: dict[int, Counts] = {}
        """The question counts per category."""

    @property
    def _client(self) -> AsyncClient:
        """The API client."""
        if self._client_ is None:
            self._client_ = AsyncClient()
        return self._client_

    async def _call(self, module: str, **params: str) -> str:
        """Call on the OpenTDB API.

        Args:
            path: The path for the API call.
            params: The parameters for the call.

        Returns:
            The text returned from the call.
        """
        try:
            response = await self._client.get(
                f"{self._BASE}/{module}.php",
                params=params,
                headers={"user-agent": self.AGENT},
            )
        except RequestError as error:
            raise self.RequestError(str(error))

        try:
            response.raise_for_status()
        except HTTPStatusError as error:
            raise self.RequestError(str(error))

        return response.text

    async def categories(self) -> list[Category]:
        """Get the list of quiz categories.

        Returns:
            The list of quiz categories.
        """
        if self._categories:
            return self._categories
        self._categories = [
            Category(**category)
            for category in loads(await self._call("api_category"))["trivia_categories"]
        ]
        return self._categories

    async def category(self, category_id: int) -> Category:
        """Get a category based on its ID.

        Args:
            category_id: The ID of the category to get.

        Returns:
            The category data.
        """
        return next(
            category
            for category in await self.categories()
            if category.id == category_id
        )

    async def questions(
        self,
        amount: int = 10,
        category: int | Category | None = None,
        difficulty: Difficulty | None = None,
        of_type: Type | None = None,
    ) -> list[Question]:
        """Get a collection of questions from the API.

        Args:
            amount: The amount of questions to get.
            category: The category of questions to get.
            difficulty: The difficulty of questions to get.
            of_type: The type of question to get.

        Returns:
            The questions.
        """

        if isinstance(category, Category):
            category = category.id

        params = {"amount": str(amount)}
        if category is not None:
            params["category"] = str(category)
        if difficulty is not None:
            params["difficulty"] = difficulty
        if of_type is not None:
            params["type"] = of_type

        response = loads(await self._call("api", **params))

        # TODO: for things like invalid or exhausted tokens, or a rate
        # limit, bounce back around, perhaps. For now though just do a
        # simple sanity check.
        Code(response.get("response_code", Code.UNKNOWN)).maybe_raise()

        return [Question(**question) for question in response["results"]]

    async def _counts(self) -> None:
        """Get the low-level counts data."""
        if self._overall_counts is not None:
            return
        counts = loads(await self._call("api_count_global"))
        self._overall_counts = Counts(
            counts["overall"]["total_num_of_questions"],
            counts["overall"]["total_num_of_pending_questions"],
            counts["overall"]["total_num_of_verified_questions"],
            counts["overall"]["total_num_of_rejected_questions"],
        )
        self._category_counts = {
            int(category): Counts(
                category_counts["total_num_of_questions"],
                category_counts["total_num_of_pending_questions"],
                category_counts["total_num_of_verified_questions"],
                category_counts["total_num_of_rejected_questions"],
            )
            for category, category_counts in counts["categories"].items()
        }

    async def overall_counts(self) -> Counts:
        """Gets the overall question counts.

        Returns:
            The overall counts from the backend.

        Note:
            Because this is a low-importance value that won't change very
            often, the value is cached for the lifetime of the client
            object.
        """
        await self._counts()
        assert self._overall_counts is not None
        return self._overall_counts

    async def category_counts(self) -> dict[int, Counts]:
        """Gets the question counts per category.

        Returns:
            The question counts per category.

        Note:
            Because this is a low-importance value that won't change very
            often, the value is cached for the lifetime of the client
            object.
        """
        await self._counts()
        return self._category_counts


### client.py ends here
